
local bg = CreateFrame("Frame")
bg:SetLayer(-10)
bg:SetBackgroundColor(0, 0.2, 0)
bg:SetAllPoints()

local function lerp(s, y, d)
  return s * (1 - d) + y * d
end



local halflife = 40
local halflife_frame = 1 - math.pow(0.5, 1 / (60 * halflife))
local exceed = 4

local map_x = 1024 * exceed
local map_y = 768 * exceed

local mapgrid = 64

local crit = CreateFrame("Frame")
crit:SetCoordinateScale(map_x / 2, map_y / 2, map_x)
crit:SetAllPoints()

local map = {}
for tx = 0, map_x / mapgrid do
  map[tx] = {}
  local tmx = map[tx]
  for ty = 0, map_y / mapgrid do
    tmx[ty] = {}
  end
end

local crits = {}
local obliterate = {}
local magnets = {}

local Critter
local Food
do
  local Critter_table = {}
  local Critter_metatable = {__index = Critter_table}
  
  Critter_table.behaviors = {}
  
  local collide_distance = 20
  local collide_distance_sqr = collide_distance * collide_distance
  local collide_weight = 0.3
  table.insert(Critter_table.behaviors, function(me, neighbors)
    local closest = nil
    local closest_dist = nil
    for _, v in ipairs(neighbors) do
      if v.dist < collide_distance_sqr and not v.player then
        if not closest_dist or neighbors.dist < closest_dist then
          closest = v
          closest_dist = neighbors.dist
        end
      end
    end
    
    if closest then
      collide_distance = math.sqrt(collide_distance)
      local dx = (me.x - closest.x) / collide_distance
      local dy = (me.y - closest.y) / collide_distance
      
      return dx * collide_weight, dy * collide_weight
    end
  end)
  
  local magnet_distance = 400
  local magnet_distance_sqr = magnet_distance * magnet_distance
  local magnet_weight = 0.2
  table.insert(Critter_table.behaviors, function(me, neighbors)
    local dx, dy = 0, 0
    
    for _, v in pairs(magnets) do
      local tx, ty = v.x - me.x, v.y - me.y
      local dd = tx * tx + ty * ty
      if dd < magnet_distance_sqr then
        dd = math.sqrt(dd)
        
        local weight = (magnet_distance - dd) / magnet_distance
        
        tx, ty = tx / dd * weight * v.strength, ty / dd * weight * v.strength
        dx, dy = dx + tx, dy + ty
      end
    end
    
    return dx, dy
  end)
  
  local cluster_age_threshold = 20
  
  local cluster_distance = 200
  local cluster_distance_sqr = cluster_distance * cluster_distance
  local cluster_weight = 1.0
  
  local function cluster(me, neighbors, friendly, enemy)
    local accum_x = 0
    local accum_y = 0
    local ct = 0
    for _, v in ipairs(neighbors) do
      if v.dist < cluster_distance_sqr then
        local weight = (cluster_age_threshold - math.abs(me.age - v.age)) / 10
        if weight > friendly then weight = friendly end
        if weight < enemy then weight = enemy end
        
        if v.player then
          ct = ct + 2
          weight = weight * 3
        end
        
        accum_x = accum_x + (v.x - me.x) * weight
        accum_y = accum_y + (v.y - me.y) * weight
        ct = ct + math.abs(weight)
      end
    end
    
    if ct > 0 then
      local acd = math.sqrt(accum_x * accum_x + accum_y * accum_y)
      return accum_x / cluster_distance * cluster_weight, accum_y / cluster_distance * cluster_weight
    end
  end
  
  table.insert(Critter_table.behaviors, function(me, neighbors)
    return cluster(me, neighbors, 0, -2)
  end)
  
  local directional_distance = 200
  local directional_distance_sqr = directional_distance * directional_distance
  local directional_weight = 0.5
  table.insert(Critter_table.behaviors, function(me, neighbors)
    local accum_x = 0
    local accum_y = 0
    local ct = 0
    local play = nil
    for _, v in ipairs(neighbors) do
      if v.dist < directional_distance_sqr then
        if (me.age - v.age) + cluster_age_threshold > 0 then
          accum_x = accum_x + v.dx
          accum_y = accum_y + v.dy
          ct = ct + 1
          if v.player then
            play = v
          end
        end
      end
    end
    
    if play then
      accum_x = accum_x + play.dx * (ct + 10)
      accum_y = accum_y + play.dy * (ct + 10)
      ct = ct * 2 + 10
      ct = ct / 2
    end
    
    if ct > 0 then
      accum_x, accum_y = accum_x / ct, accum_y / ct
      accum_x, accum_y = accum_x - me.dx, accum_y - me.dy
      return accum_x * directional_weight, accum_y * directional_weight
    end
  end)
  
  table.insert(Critter_table.behaviors, function(me, neighbors)
    return cluster(me, neighbors, 1, 0)
  end)
  
  local biggest_distance = math.max(collide_distance, cluster_distance, directional_distance)
  local biggest_distance_squared = biggest_distance * biggest_distance
  local biggest_distance_cubes = math.ceil(biggest_distance / mapgrid)
  
  local speed_change = 0.1
  
  function Critter_table:get_older()
    if math.random() < halflife_frame or self.age > 100 then self:init() end
    self.age = self.age + 100 / 60 / 60 / 2 -- 2 minutes of playtime
  end
  
  local function makeneighbors(self)
    local close = {}
    
    -- accumulate neighbors
    local tx, ty = math.floor(self.x / mapgrid), math.floor(self.y / mapgrid)
    for cx = tx - biggest_distance_cubes, tx + biggest_distance_cubes do
      local line = map[cx]
      if line then
        for cy = ty - biggest_distance_cubes, ty + biggest_distance_cubes do
          if line[cy] then
            for v in pairs(line[cy]) do
              if v ~= self then
                local dx, dy = v.x - self.x, v.y - self.y
                dx, dy = dx * dx, dy * dy
                if dx + dy < biggest_distance_squared then
                  v.dist = dx + dy
                  table.insert(close, v)
                end
              end
            end
          end
        end
      end
    end
    
    return close
  end
  
  function Critter_table:ai()
    --self:get_older()
    
    local ddx, ddy, accumed = 0, 0, 0
    local close = makeneighbors(self)
    
    -- accumulate behaviors
    for _, v in ipairs(self.behaviors) do
      local tdx, tdy = v(self, close)
      if tdx then
        local tdist = math.sqrt(tdx * tdx + tdy * tdy)
        if tdist + accumed > 1 then
          local allowed_dist = 1 - accumed
          tdx, tdy = tdx / tdist * allowed_dist, tdy / tdist * allowed_dist
          tdist = allowed_dist
        end
        
        ddx, ddy, accumed = ddx + tdx, ddy + tdy, accumed + tdist
      end
    end
    
    self.dx, self.dy = self.dx + ddx * speed_change, self.dy + ddy * speed_change
    
    -- speed approach 2.0
    local tspeed = math.sqrt(self.dx * self.dx + self.dy * self.dy)
    local apspeed = tspeed - (tspeed - self.native_speed) / 100
    self.dx, self.dy = self.dx / tspeed * apspeed, self.dy / tspeed * apspeed
    
    -- clamp speed just in case
    local tspeed = math.sqrt(self.dx * self.dx + self.dy * self.dy)
    if tspeed < 1.0 then
      self.dx, self.dy = self.dx / tspeed * 1, self.dy / tspeed * 1
    elseif tspeed > 3.0 then
      self.dx, self.dy = self.dx / tspeed * 3, self.dy / tspeed * 3
    end
  end
  
  function Critter_table:tick()
    if not map[math.floor(self.x / mapgrid)][math.floor(self.y / mapgrid)][self] then
      print("wut")
    end
    map[math.floor(self.x / mapgrid)][math.floor(self.y / mapgrid)][self] = nil
    
    self:ai()
    
    self.x = self.x + self.dx
    self.y = self.y + self.dy
    
    if self.x < 0 then self.x = self.x + map_x end
    if self.y < 0 then self.y = self.y + map_y end
    if self.x > map_x then self.x = self.x - map_x end
    if self.y > map_y then self.y = self.y - map_y end
    
    self.fram:SetPoint("CENTER", UIParent, "TOPLEFT", self.x, self.y)
    self.fram:SetCoordinateScale(0, 0, 1, math.atan2(self.dx, -self.dy))
  
    map[math.floor(self.x / mapgrid)][math.floor(self.y / mapgrid)][self] = true
  end
  function Critter_table:init()
    self.age = 0
    
    self.x = math.random() * map_x
    self.y = math.random() * map_y
    local d = math.random() * 3.14159 * 2
    local v = math.random() + 2
    self.dx = math.sin(d) * v
    self.dy = -math.cos(d) * v
    
    self.native_speed = (math.random() - 0.5) * 0.2 + 2
    
    if not self.fram then
      self.fram = CreateFrame("Frame", crit)
      local me = self -- hurf, as well as durf
      function self.fram:Tick()
        me:tick()
      end
      self.fram:SetWidth(15)
      self.fram:SetHeight(15)
      function self.fram:Draw()
        local shift = me.age / 100
        gl.Color(lerp(0.3, 0, shift), lerp(0.3, 0.7, shift), lerp(1.0, 0.0, shift))
        gl.Begin("TRIANGLE_FAN")
        gl.Vertex(0, -1)
        gl.Vertex(0.3, 0)
        gl.Vertex(0, 0.5)
        gl.Vertex(-0.3, 0)
        gl.End()
      end
    end
    
    self:tick()
  end
  
  local dista = 1
  local distribution = {}
  local dtot = 0
  for i = 1, 100 do
    table.insert(distribution, dista)
    print(dista)
    dtot = dtot + dista
    dista = dista * math.pow(0.9, 1/10)
  end
  for k in pairs(distribution) do
    distribution[k] = distribution[k] / dtot
  end
  local function mage()
    local i = math.random()
    for k, v in ipairs(distribution) do
      if i < v then return k end
      i = i - v
    end
    return 100
  end
  
  function Critter()
    local crit = setmetatable({}, Critter_metatable)
    crit:init()
    
    crit.age = mage()
    
    return crit
  end
  
  local Food_table = {}
  local Food_metatable = {__index = Food_table}
  
  local food_width = 60
  local food_range = 200
  assert(food_range <= biggest_distance)
  local food_range_square = food_range * food_range
  local food_count = 20
  function Food_table:tick()
    local close = makeneighbors(self)
    
    local ct = 0
    for _, v in ipairs(close) do
      if v.dist < food_range_square then
        ct = ct + 1
      end
    end
    
    if ct > food_count then
      -- bazoom
      print("bazoom!")
      for _, v in ipairs(close) do
        if v.dist < food_range_square then
          if v.food then v:food() end
        end
      end
      
      self:init()
    end
  end
  
  function Food_table:init()
    self.x = math.random() * map_x
    self.y = math.random() * map_y
    
    if not self.fram then
      self.fram = CreateFrame("Frame", crit)
      self.fram:SetWidth(food_width)
      self.fram:SetHeight(food_width)
      self.fram:SetBackgroundColor(0, 1, 0)
      
      local me = self
      function self.fram:Tick()
        me:tick()
      end
      function self.fram:Draw()
        gl.Color(1, 1, 1)
        gl.Begin("LINE_LOOP")
        for i = 1, 32 do
          local ang = i / 32 * 3.14159 * 2
          local x, y = math.sin(ang), math.cos(ang)
          gl.Vertex(x * food_range, y * food_range)
        end
        gl.End()
      end
    end
    
    self.fram:SetPoint("CENTER", UIParent, "TOPLEFT", self.x, self.y)
    self.fram:SetCoordinateScale(0, 0, food_width)
  end
  
  function Food()
    local self = setmetatable({}, Food_metatable)

    self:init()
    
    return self
  end
  
  local Magnet_table = {}
  local Magnet_metatable = {__index = Magnet_table}
  
  function Magnet_table:init(x, y, parity)
    self.x, self.y = x, y
    
    if not self.fram then
      self.fram = CreateFrame("Frame", crit)
      self.fram:SetWidth(food_width)
      self.fram:SetHeight(food_width)
      if parity == 1 then
        self.fram:SetBackgroundColor(0, 0, 1)
      else
        self.fram:SetBackgroundColor(1, 0, 0)
      end
      
      local me = self
      function self.fram:Draw()
        gl.Color(1, 1, 1)
        gl.Begin("LINE_LOOP")
        for i = 1, 32 do
          local ang = i / 32 * 3.14159 * 2
          local x, y = math.sin(ang), math.cos(ang)
          gl.Vertex(x * magnet_distance, y * magnet_distance)
        end
        gl.End()
      end
    end
    
    self.strength = parity
    
    self.fram:SetPoint("CENTER", UIParent, "TOPLEFT", self.x, self.y)
    self.fram:SetCoordinateScale(0, 0, food_width)
  end
  
  function Magnet(x, y, parity)
    local self = setmetatable({}, Magnet_metatable)
    
    self:init(x, y, parity)
    
    return self
  end
end

for i = 1, 200 do
  table.insert(crits, Critter())
end

table.insert(crits, Food())

crits[1].ai = function () end
crits[1].dx = 0
crits[1].dy = 0

crits[1].player = true

--crits[1].age = 100

local high = CreateFrame("Frame", crit)
high:SetWidth(200 * 2)
high:SetHeight(200 * 2)
high:SetBackgroundColor(0, 1, 0, 0.2)
high:SetPoint("CENTER", crits[1].fram, "CENTER")

local playerdir = 0
local playerspeed = 2

function tick_loop(...)
  if IsKeyDownFrame("arrow_left") then playerdir = playerdir + 0.05 end
  if IsKeyDownFrame("arrow_right") then playerdir = playerdir - 0.05 end
  if IsKeyDownFrame("arrow_up") then playerspeed = playerspeed + 0.08 end
  if IsKeyDownFrame("arrow_down") then playerspeed = playerspeed - 0.08 end
  if playerspeed < 0.5 then playerspeed = 0.5 end
  if playerspeed > 5.0 then playerspeed = 5.0 end
  crits[1].dx = math.sin(playerdir) * playerspeed
  crits[1].dy = math.cos(playerdir) * playerspeed
  
  if IsKeyDownFrame("z") then crits[1].age = 0 end
  if IsKeyDownFrame("x") then crits[1].age = 25 end
  if IsKeyDownFrame("c") then crits[1].age = 50 end
  if IsKeyDownFrame("v") then crits[1].age = 75 end
  if IsKeyDownFrame("b") then crits[1].age = 100 end
  
  local citz = {}
  for _, v in ipairs(crits) do
    if not obliterate[v] then table.insert(citz, v) end
  end
  
  crits = citz
  obliterate = {}
end
local fram = 0
function render(...)
  if IsKeyDownFrame(" ") then for i = 1, 5 do UIParent:Update() end end
  
  fram = fram + 1
  
  if fram == 600 then
    local ages = {}
    for _, v in pairs(crits) do
      print(v.dx, v.dy)
    end
    --[[
    for k, v in ipairs(ages) do
      print(k, v)
    end]]
    fram = 0
  end
    
end
function key(button, _, event)
  if event == "press" then
    local weight
    if button == "mouse_left" then weight = 1 end
    if button == "mouse_right" then weight = -1 end
    --if button == "q" then CrashHorribly() end
    if not weight then return end
    
    local x, y = GetMouseX() * exceed, GetMouseY() * exceed
    
    -- LOBSTER STICKS TO MAGNET
    table.insert(magnets, Magnet(x, y, weight))
  end
end
-- think there's more
